@@ -28,6 +28,7 @@
 #include <botan/internal/loadstor.h>
 #include <botan/internal/parsing.h>
 #include <botan/internal/stl_util.h>
+#include <botan/internal/target_info.h>
 
 #include <ctime>
 #include <iostream>
@@ -997,21 +998,6 @@ class Shim_Policy final : public Botan::TLS::Policy {
                if(std::find(supported_groups.cbegin(), supported_groups.cend(), group) != supported_groups.end()) {
                   groups.push_back(group);
                }
-
-               // Eventually the pre-standard hybrid exchange using Kyber-R3 will be
-               // retired and removed. Hence, it will likely never be part of the
-               // default `TLS::Policy::key_exchange_groups()`.
-               if(group == Botan::TLS::Group_Params::HYBRID_X25519_KYBER_768_R3_OQS) {
-                  groups.push_back(group);
-               }
-
-               // TODO: once `TLS::Policy::key_exchange_groups()` contains it by
-               //       default, remove this explicit check.
-               //
-               // See: https://github.com/randombit/botan/pull/4305
-               if(group == Botan::TLS::Group_Params::HYBRID_X25519_ML_KEM_768) {
-                  groups.push_back(group);
-               }
             }
 
             return groups;
