@@ -12,18 +12,22 @@
 
 #if defined(BOTAN_HAS_RSA)
 
+   #include <botan/numthry.h>
+   #include <botan/p11_mechanism.h>
    #include <botan/pubkey.h>
    #include <botan/rng.h>
    #include <botan/internal/blinding.h>
-   #include <botan/internal/p11_mechanism.h>
+   #include <botan/internal/mod_inv.h>
+   #include <botan/internal/monty.h>
+   #include <botan/internal/monty_exp.h>
    #include <botan/internal/pk_ops_impl.h>
 
 namespace Botan::PKCS11 {
 
 RSA_PublicKeyImportProperties::RSA_PublicKeyImportProperties(const BigInt& modulus, const BigInt& pub_exponent) :
       PublicKeyProperties(KeyType::Rsa), m_modulus(modulus), m_pub_exponent(pub_exponent) {
-   add_binary(AttributeType::Modulus, BigInt::encode(m_modulus));
-   add_binary(AttributeType::PublicExponent, BigInt::encode(m_pub_exponent));
+   add_binary(AttributeType::Modulus, m_modulus.serialize());
+   add_binary(AttributeType::PublicExponent, m_pub_exponent.serialize());
 }
 
 RSA_PublicKeyGenerationProperties::RSA_PublicKeyGenerationProperties(Ulong bits) : PublicKeyProperties(KeyType::Rsa) {
@@ -32,26 +36,26 @@ RSA_PublicKeyGenerationProperties::RSA_PublicKeyGenerationProperties(Ulong bits)
 
 PKCS11_RSA_PublicKey::PKCS11_RSA_PublicKey(Session& session, ObjectHandle handle) :
       Object(session, handle),
-      RSA_PublicKey(BigInt::decode(get_attribute_value(AttributeType::Modulus)),
-                    BigInt::decode(get_attribute_value(AttributeType::PublicExponent))) {}
+      RSA_PublicKey(BigInt::from_bytes(get_attribute_value(AttributeType::Modulus)),
+                    BigInt::from_bytes(get_attribute_value(AttributeType::PublicExponent))) {}
 
 PKCS11_RSA_PublicKey::PKCS11_RSA_PublicKey(Session& session, const RSA_PublicKeyImportProperties& pubkey_props) :
       Object(session, pubkey_props), RSA_PublicKey(pubkey_props.modulus(), pubkey_props.pub_exponent()) {}
 
 RSA_PrivateKeyImportProperties::RSA_PrivateKeyImportProperties(const BigInt& modulus, const BigInt& priv_exponent) :
       PrivateKeyProperties(KeyType::Rsa), m_modulus(modulus), m_priv_exponent(priv_exponent) {
-   add_binary(AttributeType::Modulus, BigInt::encode(m_modulus));
-   add_binary(AttributeType::PrivateExponent, BigInt::encode(m_priv_exponent));
+   add_binary(AttributeType::Modulus, m_modulus.serialize());
+   add_binary(AttributeType::PrivateExponent, m_priv_exponent.serialize());
 }
 
 PKCS11_RSA_PrivateKey::PKCS11_RSA_PrivateKey(Session& session, ObjectHandle handle) :
       Object(session, handle),
-      RSA_PublicKey(BigInt::decode(get_attribute_value(AttributeType::Modulus)),
-                    BigInt::decode(get_attribute_value(AttributeType::PublicExponent))) {}
+      RSA_PublicKey(BigInt::from_bytes(get_attribute_value(AttributeType::Modulus)),
+                    BigInt::from_bytes(get_attribute_value(AttributeType::PublicExponent))) {}
 
 PKCS11_RSA_PrivateKey::PKCS11_RSA_PrivateKey(Session& session, const RSA_PrivateKeyImportProperties& priv_key_props) :
       Object(session, priv_key_props),
-      RSA_PublicKey(priv_key_props.modulus(), BigInt::decode(get_attribute_value(AttributeType::PublicExponent))) {}
+      RSA_PublicKey(priv_key_props.modulus(), BigInt::from_bytes(get_attribute_value(AttributeType::PublicExponent))) {}
 
 PKCS11_RSA_PrivateKey::PKCS11_RSA_PrivateKey(Session& session,
                                              uint32_t bits,
@@ -76,8 +80,8 @@ PKCS11_RSA_PrivateKey::PKCS11_RSA_PrivateKey(Session& session,
 
    this->reset_handle(priv_key_handle);
 
-   BigInt n = BigInt::decode(get_attribute_value(AttributeType::Modulus));
-   BigInt e = BigInt::decode(get_attribute_value(AttributeType::PublicExponent));
+   BigInt n = BigInt::from_bytes(get_attribute_value(AttributeType::Modulus));
+   BigInt e = BigInt::from_bytes(get_attribute_value(AttributeType::PublicExponent));
    RSA_PublicKey::init(std::move(n), std::move(e));
 }
 
@@ -88,12 +92,16 @@ RSA_PrivateKey PKCS11_RSA_PrivateKey::export_key() const {
    auto d = get_attribute_value(AttributeType::PrivateExponent);
    auto n = get_attribute_value(AttributeType::Modulus);
 
-   return RSA_PrivateKey(BigInt::decode(p), BigInt::decode(q), BigInt::decode(e), BigInt::decode(d), BigInt::decode(n));
+   return RSA_PrivateKey(BigInt::from_bytes(p),
+                         BigInt::from_bytes(q),
+                         BigInt::from_bytes(e),
+                         BigInt::from_bytes(d),
+                         BigInt::from_bytes(n));
 }
 
 std::unique_ptr<Public_Key> PKCS11_RSA_PrivateKey::public_key() const {
-   return std::make_unique<RSA_PublicKey>(BigInt::decode(get_attribute_value(AttributeType::Modulus)),
-                                          BigInt::decode(get_attribute_value(AttributeType::PublicExponent)));
+   return std::make_unique<RSA_PublicKey>(BigInt::from_bytes(get_attribute_value(AttributeType::Modulus)),
+                                          BigInt::from_bytes(get_attribute_value(AttributeType::PublicExponent)));
 }
 
 secure_vector<uint8_t> PKCS11_RSA_PrivateKey::private_key_bits() const {
@@ -110,21 +118,27 @@ class PKCS11_RSA_Decryption_Operation final : public PK_Ops::Decryption {
                                       RandomNumberGenerator& rng) :
             m_key(key),
             m_mechanism(MechanismWrapper::create_rsa_crypt_mechanism(padding)),
+            m_mod_n(Barrett_Reduction::for_public_modulus(m_key.get_n())),
+            m_monty_n(std::make_shared<Montgomery_Params>(m_key.get_n(), m_mod_n)),
             m_blinder(
-               m_key.get_n(),
+               m_mod_n,
                rng,
-               [this](const BigInt& k) { return power_mod(k, m_key.get_e(), m_key.get_n()); },
-               [this](const BigInt& k) { return inverse_mod(k, m_key.get_n()); }) {
+               [this](const BigInt& k) {
+                  const size_t powm_window = 1;
+                  auto powm_m_n = monty_precompute(m_monty_n, k, powm_window, false);
+                  return monty_execute_vartime(*powm_m_n, m_key.get_e()).value();
+               },
+               [this](const BigInt& k) { return inverse_mod_rsa_public_modulus(k, m_key.get_n()); }) {
          m_bits = m_key.get_n().bits() - 1;
       }
 
       size_t plaintext_length(size_t /*ctext_len*/) const override { return m_key.get_n().bytes(); }
 
-      secure_vector<uint8_t> decrypt(uint8_t& valid_mask, const uint8_t ciphertext[], size_t ciphertext_len) override {
+      secure_vector<uint8_t> decrypt(uint8_t& valid_mask, std::span<const uint8_t> ctext) override {
          valid_mask = 0;
          m_key.module()->C_DecryptInit(m_key.session().handle(), m_mechanism.data(), m_key.handle());
 
-         std::vector<uint8_t> encrypted_data(ciphertext, ciphertext + ciphertext_len);
+         std::vector<uint8_t> encrypted_data(ctext.begin(), ctext.end());
 
          const size_t modulus_bytes = (m_key.get_n().bits() + 7) / 8;
 
@@ -132,9 +146,9 @@ class PKCS11_RSA_Decryption_Operation final : public PK_Ops::Decryption {
          const bool use_blinding = !m_mechanism.padding_size();
 
          if(use_blinding) {
-            const BigInt blinded = m_blinder.blind(BigInt::decode(encrypted_data));
+            const BigInt blinded = m_blinder.blind(BigInt::from_bytes(encrypted_data));
             // SoftHSM at least requires raw RSA inputs be == the modulus size
-            encrypted_data = unlock(BigInt::encode_1363(blinded, modulus_bytes));
+            encrypted_data = blinded.serialize(modulus_bytes);
          }
 
          secure_vector<uint8_t> decrypted_data;
@@ -142,8 +156,9 @@ class PKCS11_RSA_Decryption_Operation final : public PK_Ops::Decryption {
 
          // Unblind for RSA/RAW decryption
          if(use_blinding) {
-            const BigInt unblinded = m_blinder.unblind(BigInt::decode(decrypted_data));
-            decrypted_data = BigInt::encode_1363(unblinded, modulus_bytes);
+            const BigInt unblinded = m_blinder.unblind(BigInt::from_bytes(decrypted_data));
+            decrypted_data.resize(modulus_bytes);
+            unblinded.serialize_to(decrypted_data);
          }
 
          valid_mask = 0xFF;
@@ -153,6 +168,8 @@ class PKCS11_RSA_Decryption_Operation final : public PK_Ops::Decryption {
    private:
       const PKCS11_RSA_PrivateKey& m_key;
       MechanismWrapper m_mechanism;
+      Barrett_Reduction m_mod_n;
+      std::shared_ptr<const Montgomery_Params> m_monty_n;
       size_t m_bits = 0;
       Blinder m_blinder;
 };
@@ -168,8 +185,8 @@ class PKCS11_RSA_Decryption_Operation_Software_EME final : public PK_Ops::Decryp
 
       size_t plaintext_length(size_t ctext_len) const override { return m_raw_decryptor.plaintext_length(ctext_len); }
 
-      secure_vector<uint8_t> raw_decrypt(const uint8_t input[], size_t input_len) override {
-         return m_raw_decryptor.decrypt(input, input_len);
+      secure_vector<uint8_t> raw_decrypt(std::span<const uint8_t> input) override {
+         return m_raw_decryptor.decrypt(input);
       }
 
    private:
@@ -189,13 +206,13 @@ class PKCS11_RSA_Encryption_Operation final : public PK_Ops::Encryption {
 
       size_t max_input_bits() const override { return m_bits; }
 
-      secure_vector<uint8_t> encrypt(const uint8_t msg[], size_t msg_len, RandomNumberGenerator& /*rng*/) override {
+      std::vector<uint8_t> encrypt(std::span<const uint8_t> input, RandomNumberGenerator& /*rng*/) override {
          m_key.module()->C_EncryptInit(m_key.session().handle(), m_mechanism.data(), m_key.handle());
 
-         secure_vector<uint8_t> encrytped_data;
+         std::vector<uint8_t> encrypted_data;
          m_key.module()->C_Encrypt(
-            m_key.session().handle(), secure_vector<uint8_t>(msg, msg + msg_len), encrytped_data);
-         return encrytped_data;
+            m_key.session().handle(), secure_vector<uint8_t>(input.begin(), input.end()), encrypted_data);
+         return encrypted_data;
       }
 
    private:
@@ -211,12 +228,12 @@ class PKCS11_RSA_Signature_Operation final : public PK_Ops::Signature {
 
       size_t signature_length() const override { return m_key.get_n().bytes(); }
 
-      void update(const uint8_t msg[], size_t msg_len) override {
+      void update(std::span<const uint8_t> input) override {
          if(!m_initialized) {
             // first call to update: initialize and cache message because we can not determine yet whether a single- or multiple-part operation will be performed
             m_key.module()->C_SignInit(m_key.session().handle(), m_mechanism.data(), m_key.handle());
             m_initialized = true;
-            m_first_message = secure_vector<uint8_t>(msg, msg + msg_len);
+            m_first_message.assign(input.begin(), input.end());
             return;
          }
 
@@ -226,11 +243,11 @@ class PKCS11_RSA_Signature_Operation final : public PK_Ops::Signature {
             m_first_message.clear();
          }
 
-         m_key.module()->C_SignUpdate(m_key.session().handle(), msg, static_cast<Ulong>(msg_len));
+         m_key.module()->C_SignUpdate(m_key.session().handle(), input.data(), static_cast<Ulong>(input.size()));
       }
 
-      secure_vector<uint8_t> sign(RandomNumberGenerator& /*rng*/) override {
-         secure_vector<uint8_t> signature;
+      std::vector<uint8_t> sign(RandomNumberGenerator& /*rng*/) override {
+         std::vector<uint8_t> signature;
          if(!m_first_message.empty()) {
             // single call to update: perform single-part operation
             m_key.module()->C_Sign(m_key.session().handle(), m_first_message, signature);
@@ -326,12 +343,12 @@ class PKCS11_RSA_Verification_Operation final : public PK_Ops::Verification {
       PKCS11_RSA_Verification_Operation(const PKCS11_RSA_PublicKey& key, std::string_view padding) :
             m_key(key), m_mechanism(MechanismWrapper::create_rsa_sign_mechanism(padding)) {}
 
-      void update(const uint8_t msg[], size_t msg_len) override {
+      void update(std::span<const uint8_t> input) override {
          if(!m_initialized) {
             // first call to update: initialize and cache message because we can not determine yet whether a single- or multiple-part operation will be performed
             m_key.module()->C_VerifyInit(m_key.session().handle(), m_mechanism.data(), m_key.handle());
             m_initialized = true;
-            m_first_message = secure_vector<uint8_t>(msg, msg + msg_len);
+            m_first_message.assign(input.begin(), input.end());
             return;
          }
 
@@ -341,23 +358,24 @@ class PKCS11_RSA_Verification_Operation final : public PK_Ops::Verification {
             m_first_message.clear();
          }
 
-         m_key.module()->C_VerifyUpdate(m_key.session().handle(), msg, static_cast<Ulong>(msg_len));
+         m_key.module()->C_VerifyUpdate(m_key.session().handle(), input.data(), static_cast<Ulong>(input.size()));
       }
 
-      bool is_valid_signature(const uint8_t sig[], size_t sig_len) override {
+      bool is_valid_signature(std::span<const uint8_t> sig) override {
          ReturnValue return_value = ReturnValue::SignatureInvalid;
          if(!m_first_message.empty()) {
             // single call to update: perform single-part operation
             m_key.module()->C_Verify(m_key.session().handle(),
                                      m_first_message.data(),
                                      static_cast<Ulong>(m_first_message.size()),
-                                     sig,
-                                     static_cast<Ulong>(sig_len),
+                                     sig.data(),
+                                     static_cast<Ulong>(sig.size()),
                                      &return_value);
             m_first_message.clear();
          } else {
             // multiple calls to update (or none): finish multiple-part operation
-            m_key.module()->C_VerifyFinal(m_key.session().handle(), sig, static_cast<Ulong>(sig_len), &return_value);
+            m_key.module()->C_VerifyFinal(
+               m_key.session().handle(), sig.data(), static_cast<Ulong>(sig.size()), &return_value);
          }
          m_initialized = false;
          if(return_value != ReturnValue::OK && return_value != ReturnValue::SignatureInvalid) {
