@@ -198,7 +198,9 @@ void TLS_CBC_HMAC_AEAD_Encryption::finish_msg(secure_vector<uint8_t>& buffer, si
 
    buffer.reserve(offset + msg_size + padding_length + tag_size());
    buffer.resize(offset + msg_size);
-   copy_mem(&buffer[offset], msg().data(), msg_size);
+   if(msg_size > 0) {
+      copy_mem(&buffer[offset], msg().data(), msg_size);
+   }
 
    mac().update(assoc_data());
 
@@ -212,7 +214,9 @@ void TLS_CBC_HMAC_AEAD_Encryption::finish_msg(secure_vector<uint8_t>& buffer, si
       buffer.resize(buffer.size() + tag_size());
       mac().final(&buffer[buffer.size() - tag_size()]);
    } else {
-      mac().update(&buffer[offset], msg_size);
+      if(msg_size > 0) {
+         mac().update(&buffer[offset], msg_size);
+      }
       buffer.resize(buffer.size() + tag_size());
       mac().final(&buffer[buffer.size() - tag_size()]);
       cbc_encrypt_record(buffer, offset, padding_length);
@@ -380,9 +384,9 @@ void TLS_CBC_HMAC_AEAD_Decryption::finish_msg(secure_vector<uint8_t>& buffer, si
 
       const size_t mac_offset = enc_size;
 
-      const bool mac_ok = constant_time_compare(&record_contents[mac_offset], mac_buf.data(), tag_size());
+      const auto mac_ok = CT::is_equal(&record_contents[mac_offset], mac_buf.data(), tag_size());
 
-      if(!mac_ok) {
+      if(!mac_ok.as_bool()) {
          throw TLS_Exception(Alert::BadRecordMac, "Message authentication failure");
       }
 
@@ -441,13 +445,13 @@ void TLS_CBC_HMAC_AEAD_Decryption::finish_msg(secure_vector<uint8_t>& buffer, si
 
       const size_t mac_offset = record_len - (tag_size() + pad_size);
 
-      const bool mac_ok = constant_time_compare(&record_contents[mac_offset], mac_buf.data(), tag_size());
+      const auto mac_ok = CT::is_equal(&record_contents[mac_offset], mac_buf.data(), tag_size());
 
       const auto ok_mask = size_ok_mask & CT::Mask<uint16_t>::expand(mac_ok) & CT::Mask<uint16_t>::expand(pad_size);
 
       CT::unpoison(ok_mask);
 
-      if(ok_mask.is_set()) {
+      if(ok_mask.as_bool()) {
          buffer.insert(buffer.end(), plaintext_block, plaintext_block + plaintext_length);
       } else {
          perform_additional_compressions(record_len, pad_size);
