@@ -48,6 +48,68 @@ CertificatePathStatusCodes PKIX::check_chain(const std::vector<X509_Certificate>
 
    CertificatePathStatusCodes cert_status(cert_path.size());
 
+   // Before anything else verify the entire chain of signatures
+   for(size_t i = 0; i != cert_path.size(); ++i) {
+      std::set<Certificate_Status_Code>& status = cert_status.at(i);
+
+      const bool at_self_signed_root = (i == cert_path.size() - 1);
+
+      const X509_Certificate& subject = cert_path[i];
+      const X509_Certificate& issuer = cert_path[at_self_signed_root ? (i) : (i + 1)];
+
+      // Check the signature algorithm is known
+      if(!subject.signature_algorithm().oid().registered_oid()) {
+         status.insert(Certificate_Status_Code::SIGNATURE_ALGO_UNKNOWN);
+      } else {
+         std::unique_ptr<Public_Key> issuer_key;
+         try {
+            issuer_key = issuer.subject_public_key();
+         } catch(...) {
+            status.insert(Certificate_Status_Code::CERT_PUBKEY_INVALID);
+         }
+
+         if(issuer_key) {
+            if(issuer_key->estimated_strength() < restrictions.minimum_key_strength()) {
+               status.insert(Certificate_Status_Code::SIGNATURE_METHOD_TOO_WEAK);
+            }
+
+            const auto sig_status = subject.verify_signature(*issuer_key);
+
+            if(sig_status.first != Certificate_Status_Code::VERIFIED) {
+               status.insert(sig_status.first);
+            } else {
+               // Signature is valid, check if hash used was acceptable
+               const std::string hash_used_for_signature = sig_status.second;
+               BOTAN_ASSERT_NOMSG(!hash_used_for_signature.empty());
+               const auto& trusted_hashes = restrictions.trusted_hashes();
+
+               // Ignore untrusted hashes on self-signed roots
+               if(!trusted_hashes.empty() && !at_self_signed_root) {
+                  if(!trusted_hashes.contains(hash_used_for_signature)) {
+                     status.insert(Certificate_Status_Code::UNTRUSTED_HASH);
+                  }
+               }
+            }
+         }
+      }
+   }
+
+   // If any of the signatures were invalid, return immediately; we know the
+   // chain is invalid and signature failure is always considered the most
+   // critical result. This does mean other problems in the certificate (eg
+   // expired) will not be reported, but we'd have to assume any such data is
+   // anyway arbitrary considering we couldn't verify the signature chain
+
+   for(size_t i = 0; i != cert_path.size(); ++i) {
+      for(auto status : cert_status.at(i)) {
+         // This ignores errors relating to the key or hash being weak since
+         // these are somewhat advisory
+         if(static_cast<uint32_t>(status) >= 5000) {
+            return cert_status;
+         }
+      }
+   }
+
    if(!hostname.empty() && !cert_path[0].matches_dns_name(hostname)) {
       cert_status[0].insert(Certificate_Status_Code::CERT_NAME_NOMATCH);
    }
@@ -83,6 +145,7 @@ CertificatePathStatusCodes PKIX::check_chain(const std::vector<X509_Certificate>
          status.insert(Certificate_Status_Code::CHAIN_LACKS_TRUST_ROOT);
       }
 
+      // This should never happen; it indicates a bug in path building
       if(subject.issuer_dn() != issuer.subject_dn()) {
          status.insert(Certificate_Status_Code::CHAIN_NAME_MISMATCH);
       }
@@ -102,13 +165,24 @@ CertificatePathStatusCodes PKIX::check_chain(const std::vector<X509_Certificate>
          }
       }
 
+      // Only warn, if trusted root is not in time range if configured this way
+      const bool is_trusted_root_and_time_ignored =
+         restrictions.ignore_trusted_root_time_range() && at_self_signed_root;
       // Check all certs for valid time range
       if(validation_time < subject.not_before()) {
-         status.insert(Certificate_Status_Code::CERT_NOT_YET_VALID);
+         if(is_trusted_root_and_time_ignored) {
+            status.insert(Certificate_Status_Code::TRUSTED_CERT_NOT_YET_VALID);  // only warn
+         } else {
+            status.insert(Certificate_Status_Code::CERT_NOT_YET_VALID);
+         }
       }
 
       if(validation_time > subject.not_after()) {
-         status.insert(Certificate_Status_Code::CERT_HAS_EXPIRED);
+         if(is_trusted_root_and_time_ignored) {
+            status.insert(Certificate_Status_Code::TRUSTED_CERT_HAS_EXPIRED);  // only warn
+         } else {
+            status.insert(Certificate_Status_Code::CERT_HAS_EXPIRED);
+         }
       }
 
       // Check issuer constraints
@@ -116,39 +190,6 @@ CertificatePathStatusCodes PKIX::check_chain(const std::vector<X509_Certificate>
          status.insert(Certificate_Status_Code::CA_CERT_NOT_FOR_CERT_ISSUER);
       }
 
-      auto issuer_key = issuer.subject_public_key();
-
-      // Check the signature algorithm is known
-      if(!subject.signature_algorithm().oid().registered_oid()) {
-         status.insert(Certificate_Status_Code::SIGNATURE_ALGO_UNKNOWN);
-      } else {
-         // only perform the following checks if the signature algorithm is known
-         if(!issuer_key) {
-            status.insert(Certificate_Status_Code::CERT_PUBKEY_INVALID);
-         } else {
-            const auto sig_status = subject.verify_signature(*issuer_key);
-
-            if(sig_status.first == Certificate_Status_Code::VERIFIED) {
-               const std::string hash_used_for_signature = sig_status.second;
-               BOTAN_ASSERT_NOMSG(!hash_used_for_signature.empty());
-               const auto& trusted_hashes = restrictions.trusted_hashes();
-
-               // Ignore untrusted hashes on self-signed roots
-               if(!trusted_hashes.empty() && !at_self_signed_root) {
-                  if(!trusted_hashes.contains(hash_used_for_signature)) {
-                     status.insert(Certificate_Status_Code::UNTRUSTED_HASH);
-                  }
-               }
-            } else {
-               status.insert(sig_status.first);
-            }
-
-            if(issuer_key->estimated_strength() < restrictions.minimum_key_strength()) {
-               status.insert(Certificate_Status_Code::SIGNATURE_METHOD_TOO_WEAK);
-            }
-         }
-      }
-
       // Check cert extensions
 
       if(subject.x509_version() == 1) {
@@ -343,7 +384,7 @@ CertificatePathStatusCodes PKIX::check_crl(const std::vector<X509_Certificate>&
             status.insert(Certificate_Status_Code::CRL_NOT_YET_VALID);
          }
 
-         if(validation_time > crls[i]->next_update()) {
+         if(crls[i]->next_update().time_is_set() && validation_time > crls[i]->next_update()) {
             status.insert(Certificate_Status_Code::CRL_HAS_EXPIRED);
          }
 
@@ -358,9 +399,11 @@ CertificatePathStatusCodes PKIX::check_crl(const std::vector<X509_Certificate>&
             status.insert(Certificate_Status_Code::CERT_IS_REVOKED);
          }
 
-         std::string dp = subject.crl_distribution_point();
+         const auto dp = subject.crl_distribution_points();
          if(!dp.empty()) {
-            if(dp != crls[i]->crl_issuing_distribution_point()) {
+            const auto crl_idp = crls[i]->crl_issuing_distribution_point();
+
+            if(std::find(dp.begin(), dp.end(), crl_idp) == dp.end()) {
                status.insert(Certificate_Status_Code::NO_MATCHING_CRLDP);
             }
          }
@@ -444,7 +487,7 @@ CertificatePathStatusCodes PKIX::check_ocsp_online(const std::vector<X509_Certif
          }));
       } else {
          ocsp_response_futures.emplace_back(std::async(std::launch::async, [&]() -> std::optional<OCSP::Response> {
-            OCSP::Request req(issuer, BigInt::decode(subject.serial_number()));
+            OCSP::Request req(issuer, BigInt::from_bytes(subject.serial_number()));
 
             HTTP::Response http;
             try {
@@ -607,8 +650,8 @@ Certificate_Status_Code PKIX::build_certificate_path(std::vector<X509_Certificat
 
       const std::string fprint = issuer->fingerprint("SHA-256");
 
-      if(certs_seen.contains(fprint))  // already seen?
-      {
+      if(certs_seen.contains(fprint)) {
+         // we already saw this certificate -> loop
          return Certificate_Status_Code::CERT_CHAIN_LOOP;
       }
 
@@ -796,10 +839,10 @@ void PKIX::merge_revocation_status(CertificatePathStatusCodes& chain_status,
 
       if(i < ocsp.size() && !ocsp[i].empty()) {
          for(auto&& code : ocsp[i]) {
+            // NO_REVOCATION_URL and OCSP_SERVER_NOT_AVAILABLE are softfail
             if(code == Certificate_Status_Code::OCSP_RESPONSE_GOOD ||
-               code == Certificate_Status_Code::OCSP_NO_REVOCATION_URL ||   // softfail
-               code == Certificate_Status_Code::OCSP_SERVER_NOT_AVAILABLE)  // softfail
-            {
+               code == Certificate_Status_Code::OCSP_NO_REVOCATION_URL ||
+               code == Certificate_Status_Code::OCSP_SERVER_NOT_AVAILABLE) {
                had_ocsp = true;
             }
 
@@ -945,12 +988,14 @@ Path_Validation_Restrictions::Path_Validation_Restrictions(bool require_rev,
                                                            size_t key_strength,
                                                            bool ocsp_intermediates,
                                                            std::chrono::seconds max_ocsp_age,
-                                                           std::unique_ptr<Certificate_Store> trusted_ocsp_responders) :
+                                                           std::unique_ptr<Certificate_Store> trusted_ocsp_responders,
+                                                           bool ignore_trusted_root_time_range) :
       m_require_revocation_information(require_rev),
       m_ocsp_all_intermediates(ocsp_intermediates),
       m_minimum_key_strength(key_strength),
       m_max_ocsp_age(max_ocsp_age),
-      m_trusted_ocsp_responders(std::move(trusted_ocsp_responders)) {
+      m_trusted_ocsp_responders(std::move(trusted_ocsp_responders)),
+      m_ignore_trusted_root_time_range(ignore_trusted_root_time_range) {
    if(key_strength <= 80) {
       m_trusted_hashes.insert("SHA-1");
    }
@@ -959,7 +1004,8 @@ Path_Validation_Restrictions::Path_Validation_Restrictions(bool require_rev,
    m_trusted_hashes.insert("SHA-256");
    m_trusted_hashes.insert("SHA-384");
    m_trusted_hashes.insert("SHA-512");
-   m_trusted_hashes.insert("SHAKE-256(512)");
+   m_trusted_hashes.insert("SHAKE-256(512)");  // Dilithium/ML-DSA
+   m_trusted_hashes.insert("SHAKE-256(912)");  // Ed448
 }
 
 namespace {
@@ -983,7 +1029,7 @@ Path_Validation_Result::Path_Validation_Result(CertificatePathStatusCodes status
                                                std::vector<X509_Certificate>&& cert_chain) :
       m_all_status(std::move(status)),
       m_warnings(find_warnings(m_all_status)),
-      m_cert_path(cert_chain),
+      m_cert_path(std::move(cert_chain)),
       m_overall(PKIX::overall_status(m_all_status)) {}
 
 const X509_Certificate& Path_Validation_Result::trust_root() const {
