@@ -1,5 +1,6 @@
 /*
 * (C) 2009,2015,2016 Jack Lloyd
+* (C) 2024           Fabian Albert, René Meusel -  Rohde & Schwarz Cybersecurity
 *
 * Botan is released under the Simplified BSD License (see license.txt)
 */
@@ -10,10 +11,15 @@
    #include "test_rng.h"
    #include <botan/bigint.h>
    #include <botan/numthry.h>
-   #include <botan/reducer.h>
+   #include <botan/internal/barrett.h>
+   #include <botan/internal/ct_utils.h>
    #include <botan/internal/divide.h>
+   #include <botan/internal/fmt.h>
+   #include <botan/internal/mod_inv.h>
+   #include <botan/internal/mp_core.h>
    #include <botan/internal/parsing.h>
    #include <botan/internal/primality.h>
+   #include <botan/internal/stl_util.h>
 #endif
 
 namespace Botan_Tests {
@@ -78,36 +84,36 @@ class BigInt_Unit_Tests final : public Test {
       static Test::Result test_random_prime() {
          Test::Result result("BigInt prime generation");
 
-         result.test_throws("Invalid bit size", "random_prime: Can't make a prime of 0 bits", []() {
-            Botan::random_prime(Test::rng(), 0);
-         });
-         result.test_throws("Invalid bit size", "random_prime: Can't make a prime of 1 bits", []() {
-            Botan::random_prime(Test::rng(), 1);
-         });
-         result.test_throws("Invalid arg", "random_prime Invalid value for equiv/modulo", []() {
-            Botan::random_prime(Test::rng(), 2, 1, 0, 2);
+         auto rng = Test::new_rng("random_prime");
+
+         result.test_throws(
+            "Invalid bit size", "random_prime: Can't make a prime of 0 bits", [&]() { Botan::random_prime(*rng, 0); });
+         result.test_throws(
+            "Invalid bit size", "random_prime: Can't make a prime of 1 bits", [&]() { Botan::random_prime(*rng, 1); });
+         result.test_throws("Invalid arg", "random_prime Invalid value for equiv/modulo", [&]() {
+            Botan::random_prime(*rng, 2, 1, 0, 2);
          });
 
-         BigInt p = Botan::random_prime(Test::rng(), 2);
+         BigInt p = Botan::random_prime(*rng, 2);
          result.confirm("Only two 2-bit primes", p == 2 || p == 3);
 
-         p = Botan::random_prime(Test::rng(), 3);
+         p = Botan::random_prime(*rng, 3);
          result.confirm("Only two 3-bit primes", p == 5 || p == 7);
 
-         p = Botan::random_prime(Test::rng(), 4);
+         p = Botan::random_prime(*rng, 4);
          result.confirm("Only two 4-bit primes", p == 11 || p == 13);
 
          for(size_t bits = 5; bits <= 32; ++bits) {
-            p = Botan::random_prime(Test::rng(), bits);
+            p = Botan::random_prime(*rng, bits);
             result.test_eq("Expected bit size", p.bits(), bits);
-            result.test_eq("P is prime", Botan::is_prime(p, Test::rng()), true);
+            result.test_eq("P is prime", Botan::is_prime(p, *rng), true);
          }
 
          const size_t safe_prime_bits = 65;
-         const BigInt safe_prime = Botan::random_safe_prime(Test::rng(), safe_prime_bits);
+         const BigInt safe_prime = Botan::random_safe_prime(*rng, safe_prime_bits);
          result.test_eq("Safe prime size", safe_prime.bits(), safe_prime_bits);
-         result.confirm("P is prime", Botan::is_prime(safe_prime, Test::rng()));
-         result.confirm("(P-1)/2 is prime", Botan::is_prime((safe_prime - 1) / 2, Test::rng()));
+         result.confirm("P is prime", Botan::is_prime(safe_prime, *rng));
+         result.confirm("(P-1)/2 is prime", Botan::is_prime((safe_prime - 1) / 2, *rng));
 
          return result;
       }
@@ -115,20 +121,19 @@ class BigInt_Unit_Tests final : public Test {
       static Test::Result test_encode() {
          Test::Result result("BigInt encoding functions");
 
-         const BigInt n1(0xffff);
-         const BigInt n2(1023);
+         const auto n1 = Botan::BigInt::from_u64(0xffff);
+         const auto n2 = Botan::BigInt::from_u64(1023);
 
-         Botan::secure_vector<uint8_t> encoded_n1 = BigInt::encode_1363(n1, 256);
-         Botan::secure_vector<uint8_t> encoded_n2 = BigInt::encode_1363(n2, 256);
-         Botan::secure_vector<uint8_t> expected = encoded_n1;
-         expected += encoded_n2;
+         const auto encoded_n1 = n1.serialize(256);
+         const auto encoded_n2 = n2.serialize(256);
+         const auto expected = Botan::concat(encoded_n1, encoded_n2);
 
-         Botan::secure_vector<uint8_t> encoded_n1_n2 = BigInt::encode_fixed_length_int_pair(n1, n2, 256);
+         const auto encoded_n1_n2 = BigInt::encode_fixed_length_int_pair(n1, n2, 256);
          result.test_eq("encode_fixed_length_int_pair", encoded_n1_n2, expected);
 
          for(size_t i = 0; i < 256 - n1.bytes(); ++i) {
             if(encoded_n1[i] != 0) {
-               result.test_failure("encode_1363", "no zero byte");
+               result.test_failure("BigInt::serialize", "no zero byte");
             }
          }
 
@@ -140,7 +145,9 @@ class BigInt_Unit_Tests final : public Test {
 
          const size_t rbits = 1024;
 
-         const Botan::BigInt r(Test::rng(), rbits);
+         auto rng = Test::new_rng("get_substring");
+
+         const Botan::BigInt r(*rng, rbits);
 
          for(size_t wlen = 1; wlen <= 32; ++wlen) {
             for(size_t offset = 0; offset != rbits + 64; ++offset) {
@@ -394,6 +401,43 @@ class BigInt_Div_Test final : public Text_Based_Test {
 
 BOTAN_REGISTER_TEST("math", "bn_div", BigInt_Div_Test);
 
+class BigInt_DivPow2k_Test final : public Test {
+   public:
+      std::vector<Test::Result> run() override {
+         Test::Result result("BigInt ct_divide_pow2k");
+
+         for(size_t k = 2; k != 128; ++k) {
+            auto div1 = Botan::ct_divide_pow2k(k, 1);
+            result.test_eq("ct_divide_pow2k div 1", div1, Botan::BigInt::power_of_2(k));
+
+            auto div2 = Botan::ct_divide_pow2k(k, 2);
+            result.test_eq("ct_divide_pow2k div 2", div2, Botan::BigInt::power_of_2(k - 1));
+
+            auto div4 = Botan::ct_divide_pow2k(k, 4);
+            result.test_eq("ct_divide_pow2k div 4", div4, Botan::BigInt::power_of_2(k - 2));
+         }
+
+         for(size_t k = 4; k != 512; ++k) {
+            const BigInt pow2k = BigInt::power_of_2(k);
+
+            for(size_t y_bits = k / 2; y_bits <= (k + 2); ++y_bits) {
+               const BigInt y(rng(), y_bits, false);
+               if(y.is_zero()) {
+                  continue;
+               }
+               const BigInt ct_pow2k = ct_divide_pow2k(k, y);
+               const BigInt ref = BigInt::power_of_2(k) / y;
+
+               result.test_eq("ct_divide_pow2k matches Knuth division", ct_pow2k, ref);
+            }
+         }
+
+         return {result};
+      }
+};
+
+BOTAN_REGISTER_TEST("math", "bn_div_pow2k", BigInt_DivPow2k_Test);
+
 class BigInt_Mod_Test final : public Text_Based_Test {
    public:
       BigInt_Mod_Test() : Text_Based_Test("bn/mod.vec", "In1,In2,Output") {}
@@ -411,8 +455,13 @@ class BigInt_Mod_Test final : public Text_Based_Test {
          e %= b;
          result.test_eq("a %= b", e, expected);
 
-         const Botan::Modular_Reducer mod_b(b);
-         result.test_eq("Barrett", mod_b.reduce(a), expected);
+         if(a.is_positive() && a < (b * b)) {
+            auto mod_b_pub = Botan::Barrett_Reduction::for_public_modulus(b);
+            result.test_eq("Barrett public", mod_b_pub.reduce(a), expected);
+
+            auto mod_b_sec = Botan::Barrett_Reduction::for_secret_modulus(b);
+            result.test_eq("Barrett secret", mod_b_sec.reduce(a), expected);
+         }
 
          // if b fits into a Botan::word test %= operator for words
          if(b.sig_words() == 1) {
@@ -440,6 +489,55 @@ class BigInt_Mod_Test final : public Text_Based_Test {
 
 BOTAN_REGISTER_TEST("math", "bn_mod", BigInt_Mod_Test);
 
+class Barrett_Redc_Test final : public Test {
+   public:
+      std::vector<Test::Result> run() override {
+         Test::Result result("Barrett reduction");
+
+         for(size_t t = 0; t != 10000; ++t) {
+            const auto mod = [&]() {
+               if(t <= 16) {
+                  return BigInt::from_u64(t) + 1;
+               } else {
+                  const size_t bits = (t / 4) + 2;
+
+                  if(t % 4 == 0) {
+                     return BigInt::power_of_2(bits);
+                  } else if(t % 4 == 1) {
+                     return BigInt::power_of_2(bits) - 1;
+                  } else if(t % 4 == 2) {
+                     return BigInt::power_of_2(bits) + 1;
+                  } else {
+                     Botan::BigInt b;
+                     b.randomize(rng(), bits, true);
+                     return b;
+                  }
+               }
+            }();
+
+            const size_t mod_bits = mod.bits();
+            auto barrett = Botan::Barrett_Reduction::for_public_modulus(mod);
+
+            for(size_t i = 0; i != 10; ++i) {
+               const auto input = [&]() {
+                  Botan::BigInt b;
+                  b.randomize(rng(), 2 * mod_bits, false);
+                  return b;
+               }();
+
+               const auto reduced_ref = input % mod;
+               const auto reduced_barrett = barrett.reduce(input);
+
+               result.test_eq("Barrett reduction matches variable time division", reduced_barrett, reduced_ref);
+            }
+         }
+
+         return {result};
+      }
+};
+
+BOTAN_REGISTER_TEST("math", "barrett_redc", Barrett_Redc_Test);
+
 class BigInt_GCD_Test final : public Text_Based_Test {
    public:
       BigInt_GCD_Test() : Text_Based_Test("bn/gcd.vec", "X,Y,GCD") {}
@@ -536,6 +634,36 @@ class BigInt_Rshift_Test final : public Text_Based_Test {
 
 BOTAN_REGISTER_TEST("math", "bn_rshift", BigInt_Rshift_Test);
 
+Test::Result test_const_time_left_shift() {
+   Test::Result result("BigInt const time shift");
+   const size_t bits = Test::run_long_tests() ? 4096 : 2048;
+
+   auto rng = Test::new_rng("const_time_left_shift");
+
+   result.start_timer();
+
+   Botan::BigInt a = Botan::BigInt::with_capacity(bits / sizeof(Botan::word));
+   for(size_t i = 0; i < bits; ++i) {
+      if(rng->next_byte() & 1) {
+         a.set_bit(i);
+      }
+   }
+
+   for(size_t i = 0; i < bits; ++i) {
+      auto ct = a;
+      auto chk = a;
+      Botan::CT::poison(ct);
+      ct.ct_shift_left(i);
+      Botan::CT::unpoison(ct);
+      chk <<= i;
+      result.test_eq(Botan::fmt("ct << {}", i), ct, chk);
+   }
+
+   result.end_timer();
+
+   return result;
+}
+
 class BigInt_Powmod_Test final : public Text_Based_Test {
    public:
       BigInt_Powmod_Test() : Text_Based_Test("bn/powmod.vec", "Base,Exponent,Modulus,Output") {}
@@ -568,7 +696,7 @@ class BigInt_IsPrime_Test final : public Text_Based_Test {
          const bool is_prime = (header == "Prime");
 
          Test::Result result("BigInt Test " + header);
-         result.test_eq("is_prime", Botan::is_prime(value, Test::rng()), is_prime);
+         result.test_eq("is_prime", Botan::is_prime(value, this->rng()), is_prime);
 
          return result;
       }
@@ -630,19 +758,23 @@ class BigInt_InvMod_Test final : public Text_Based_Test {
          const Botan::BigInt mod = vars.get_req_bn("Modulus");
          const Botan::BigInt expected = vars.get_req_bn("Output");
 
-         const Botan::BigInt a_inv = Botan::inverse_mod(a, mod);
+         result.test_eq("inverse_mod", Botan::inverse_mod(a, mod), expected);
 
-         result.test_eq("inverse_mod", a_inv, expected);
-
-         if(a_inv > 1) {
-            result.test_eq("inverse ok", (a * a_inv) % mod, 1);
-         }
-         /*
-         else if((a % mod) > 0)
-            {
-            result.confirm("no inverse with gcd > 1", gcd(a, mod) > 1);
+         if(a < mod && a > 0 && a < mod) {
+            auto g = Botan::inverse_mod_general(a, mod);
+            if(g.has_value()) {
+               result.test_eq("inverse_mod_general", g.value(), expected);
+               result.test_eq("inverse works", ((g.value() * a) % mod), BigInt::one());
+            } else {
+               result.confirm("inverse_mod_general", expected.is_zero());
             }
-         */
+
+            if(Botan::is_prime(mod, rng()) && mod != 2) {
+               BOTAN_ASSERT_NOMSG(expected > 0);
+               result.test_eq("inverse_mod_secret_prime", Botan::inverse_mod_secret_prime(a, mod), expected);
+               result.test_eq("inverse_mod_public_prime", Botan::inverse_mod_public_prime(a, mod), expected);
+            }
+         }
 
          return result;
       }
@@ -676,7 +808,7 @@ BOTAN_REGISTER_TEST("math", "bn_rand", BigInt_Rand_Test);
 class Lucas_Primality_Test final : public Test {
    public:
       std::vector<Test::Result> run() override {
-         const uint32_t lucas_max = (Test::run_long_tests() ? 100000 : 6000);
+         const uint32_t lucas_max = (Test::run_long_tests() ? 100000 : 10000) + 1;
 
          // OEIS A217120
          std::set<uint32_t> lucas_pp{
@@ -689,9 +821,9 @@ class Lucas_Primality_Test final : public Test {
          Test::Result result("Lucas primality test");
 
          for(uint32_t i = 3; i <= lucas_max; i += 2) {
-            Botan::Modular_Reducer mod_i(i);
+            auto mod_i = Botan::Barrett_Reduction::for_public_modulus(i);
             const bool passes_lucas = Botan::is_lucas_probable_prime(i, mod_i);
-            const bool is_prime = Botan::is_prime(i, Test::rng());
+            const bool is_prime = Botan::is_prime(i, this->rng());
 
             const bool is_lucas_pp = (is_prime == false && passes_lucas == true);
 
@@ -708,6 +840,58 @@ class Lucas_Primality_Test final : public Test {
 
 BOTAN_REGISTER_TEST("math", "bn_lucas", Lucas_Primality_Test);
 
+class RSA_Compute_Exp_Test : public Test {
+   public:
+      std::vector<Test::Result> run() override {
+         const size_t iter = 4000;
+
+         Test::Result result("RSA compute exponent");
+
+         const auto e = Botan::BigInt::from_u64(65537);
+
+         /*
+         * Rather than create a fresh p/q for each iteration this test creates
+         * a pool of primes then selects 2 at random as p/q
+         */
+
+         const auto random_primes = [&]() {
+            std::vector<Botan::BigInt> rp;
+            for(size_t i = 0; i != iter / 10; ++i) {
+               size_t bits = (128 + (i % 1024)) % 4096;
+               auto p = Botan::random_prime(rng(), bits);
+               if(gcd(p - 1, e) == 1) {
+                  rp.push_back(p);
+               }
+            }
+            return rp;
+         }();
+
+         for(size_t i = 0; i != iter; ++i) {
+            const size_t p_idx = random_index(rng(), random_primes.size());
+            const size_t q_idx = random_index(rng(), random_primes.size());
+
+            if(p_idx == q_idx) {
+               continue;
+            }
+
+            const auto& p = random_primes[p_idx];
+            const auto& q = random_primes[q_idx];
+
+            auto phi_n = lcm(p - 1, q - 1);
+
+            auto d = Botan::compute_rsa_secret_exponent(e, phi_n, p, q);
+
+            auto one = (e * d) % phi_n;
+
+            result.test_eq("compute_rsa_secret_exponent returned inverse", (e * d) % phi_n, Botan::BigInt::one());
+         }
+
+         return {result};
+      }
+};
+
+BOTAN_REGISTER_TEST("math", "rsa_compute_d", RSA_Compute_Exp_Test);
+
 class DSA_ParamGen_Test final : public Text_Based_Test {
    public:
       DSA_ParamGen_Test() : Text_Based_Test("bn/dsa_gen.vec", "P,Q,Counter,Seed") {}
@@ -732,7 +916,7 @@ class DSA_ParamGen_Test final : public Text_Based_Test {
 
          try {
             Botan::BigInt gen_P, gen_Q;
-            if(Botan::generate_dsa_primes(Test::rng(), gen_P, gen_Q, p_bits, q_bits, seed, offset)) {
+            if(Botan::generate_dsa_primes(this->rng(), gen_P, gen_Q, p_bits, q_bits, seed, offset)) {
                result.test_eq("P", gen_P, exp_P);
                result.test_eq("Q", gen_Q, exp_Q);
             } else {
@@ -746,6 +930,58 @@ class DSA_ParamGen_Test final : public Text_Based_Test {
 
 BOTAN_REGISTER_TEST("math", "dsa_param", DSA_ParamGen_Test);
 
+std::vector<Test::Result> test_bigint_serialization() {
+   auto rng = Test::new_rng("test_bigint_serialization");
+
+   return {
+      CHECK("BigInt binary serialization",
+            [](Test::Result& res) {
+               Botan::BigInt a(0x1234567890ABCDEF);
+               auto enc = a.serialize();
+               res.test_eq("BigInt::serialize", enc, Botan::hex_decode("1234567890ABCDEF"));
+
+               auto enc10 = a.serialize(10);
+               res.test_eq("BigInt::serialize", enc10, Botan::hex_decode("00001234567890ABCDEF"));
+
+               res.test_throws("BigInt::serialize rejects short output", [&]() { a.serialize(7); });
+            }),
+
+      CHECK("BigInt truncated/padded binary serialization",
+            [&](Test::Result& res) {
+               Botan::BigInt a(0xFEDCBA9876543210);
+
+               std::vector<uint8_t> enc1(a.bytes() - 1);
+               a.binary_encode(enc1.data(), enc1.size());
+               res.test_eq("BigInt::binary_encode", enc1, Botan::hex_decode("DCBA9876543210"));
+
+               std::vector<uint8_t> enc2(a.bytes() - 3);
+               a.binary_encode(enc2.data(), enc2.size());
+               res.test_eq("BigInt::binary_encode", enc2, Botan::hex_decode("9876543210"));
+
+               std::vector<uint8_t> enc3(a.bytes() + 1);
+               a.binary_encode(enc3.data(), enc3.size());
+               res.test_eq("BigInt::binary_encode", enc3, Botan::hex_decode("00FEDCBA9876543210"));
+
+               // make sure that the padding is actually written
+               std::vector<uint8_t> enc4(a.bytes() + 3);
+               rng->randomize(enc4);
+               a.binary_encode(enc4.data(), enc4.size());
+               res.test_eq("BigInt::binary_encode", enc4, Botan::hex_decode("000000FEDCBA9876543210"));
+
+               Botan::BigInt b(Botan::hex_decode("FEDCBA9876543210BAADC0FFEE"));
+
+               std::vector<uint8_t> enc5(b.bytes() + 12);
+               rng->randomize(enc5);
+               b.binary_encode(enc5.data(), enc5.size());
+               res.test_eq("BigInt::binary_encode",
+                           enc5,
+                           Botan::hex_decode("000000000000000000000000FEDCBA9876543210BAADC0FFEE"));
+            }),
+   };
+}
+
+BOTAN_REGISTER_TEST_FN("math", "bignum_auxiliary", test_const_time_left_shift, test_bigint_serialization);
+
 #endif
 
 }  // namespace
