@@ -17,13 +17,15 @@
    #include <botan/hash.h>
    #include <botan/pk_algs.h>
    #include <botan/pubkey.h>
+   #include <botan/internal/fmt.h>
+   #include <algorithm>
 
    #include "test_pubkey.h"
 #endif
 
 namespace Botan_Tests {
 
-#if defined(BOTAN_HAS_DILITHIUM_COMMON) && defined(BOTAN_HAS_AES)
+#if defined(BOTAN_HAS_DILITHIUM_COMMON) && defined(BOTAN_HAS_AES) && defined(BOTAN_HAS_SHA3)
 
 template <typename DerivedT>
 class Dilithium_KAT_Tests : public Text_Based_Test {
@@ -66,7 +68,7 @@ class Dilithium_KAT_Tests : public Text_Based_Test {
          verifier.update(ref_msg.data(), ref_msg.size());
          result.confirm("signature verifies", verifier.check_signature(signature.data(), signature.size()));
 
-         // test validating incorrect wrong signagture
+         // test validating incorrect wrong signature
          auto mutated_signature = Test::mutate_vec(signature, this->rng());
          result.confirm("invalid signature rejected",
                         !verifier.check_signature(mutated_signature.data(), mutated_signature.size()));
@@ -86,7 +88,17 @@ class Dilithium_KAT_Tests : public Text_Based_Test {
             constexpr static auto mode = Botan::DilithiumMode::Dilithium##m;             \
             constexpr static auto sign_param = #rand;                                    \
       };                                                                                 \
-      BOTAN_REGISTER_TEST("dilithium", "dilithium_kat_" #m "_" #rand, DILITHIUM##m##rand)
+      BOTAN_REGISTER_TEST("pubkey", "dilithium_kat_" #m "_" #rand, DILITHIUM##m##rand)
+
+   // NOLINTNEXTLINE(*-macro-usage)
+   #define REGISTER_ML_DSA_KAT_TEST(m, rand)                                          \
+      class ML_DSA##m##rand final : public Dilithium_KAT_Tests<ML_DSA##m##rand> {     \
+         public:                                                                      \
+            constexpr static auto test_vector = "pubkey/ml-dsa-" #m "_" #rand ".vec"; \
+            constexpr static auto mode = Botan::DilithiumMode::ML_DSA_##m;            \
+            constexpr static auto sign_param = #rand;                                 \
+      };                                                                              \
+      BOTAN_REGISTER_TEST("pubkey", "ml-dsa_kat_" #m "_" #rand, ML_DSA##m##rand)
 
    #if defined(BOTAN_HAS_DILITHIUM)
 REGISTER_DILITHIUM_KAT_TEST(4x4, Deterministic);
@@ -106,10 +118,24 @@ REGISTER_DILITHIUM_KAT_TEST(6x5_AES, Randomized);
 REGISTER_DILITHIUM_KAT_TEST(8x7_AES, Randomized);
    #endif
 
+   #if defined(BOTAN_HAS_ML_DSA)
+REGISTER_ML_DSA_KAT_TEST(4x4, Deterministic);
+REGISTER_ML_DSA_KAT_TEST(6x5, Deterministic);
+REGISTER_ML_DSA_KAT_TEST(8x7, Deterministic);
+REGISTER_ML_DSA_KAT_TEST(4x4, Randomized);
+REGISTER_ML_DSA_KAT_TEST(6x5, Randomized);
+REGISTER_ML_DSA_KAT_TEST(8x7, Randomized);
+   #endif
+
 class DilithiumRoundtripTests final : public Test {
    public:
-      static Test::Result run_roundtrip(const char* test_name, Botan::DilithiumMode mode, bool randomized) {
+      static Test::Result run_roundtrip(
+         const char* test_name, Botan::DilithiumMode mode, bool randomized, size_t strength, size_t psid) {
          Test::Result result(test_name);
+         if(!mode.is_available()) {
+            result.note_missing(mode.to_string());
+            return result;
+         }
 
          auto rng = Test::new_rng(test_name);
 
@@ -131,6 +157,11 @@ class DilithiumRoundtripTests final : public Test {
          Botan::Dilithium_PrivateKey priv_key(*rng, mode);
          const Botan::Dilithium_PublicKey& pub_key = priv_key;
 
+         result.test_eq("key strength", priv_key.estimated_strength(), strength);
+         result.test_eq("key length", priv_key.key_length(), psid);
+         result.test_eq("key strength", pub_key.estimated_strength(), strength);
+         result.test_eq("key length", pub_key.key_length(), psid);
+
          const auto sig_before_codec = sign(priv_key, msgvec);
 
          const auto priv_key_encoded = priv_key.private_key_bits();
@@ -180,46 +211,68 @@ class DilithiumRoundtripTests final : public Test {
       }
 
       std::vector<Test::Result> run() override {
-         std::vector<Test::Result> results;
-
-   #if defined(BOTAN_HAS_DILITHIUM)
-         results.push_back(run_roundtrip("Dilithium_4x4_Common", Botan::DilithiumMode::Dilithium4x4, false));
-         results.push_back(run_roundtrip("Dilithium_6x5_Common", Botan::DilithiumMode::Dilithium6x5, false));
-         results.push_back(run_roundtrip("Dilithium_8x7_Common", Botan::DilithiumMode::Dilithium8x7, false));
-         results.push_back(run_roundtrip("Dilithium_4x4_Common_Randomized", Botan::DilithiumMode::Dilithium4x4, true));
-         results.push_back(run_roundtrip("Dilithium_6x5_Common_Randomized", Botan::DilithiumMode::Dilithium6x5, true));
-         results.push_back(run_roundtrip("Dilithium_8x7_Common_Randomized", Botan::DilithiumMode::Dilithium8x7, true));
-   #endif
-
-   #if defined(BOTAN_HAS_DILITHIUM_AES)
-         results.push_back(run_roundtrip("Dilithium_4x4_AES", Botan::DilithiumMode::Dilithium4x4_AES, false));
-         results.push_back(run_roundtrip("Dilithium_6x5_AES", Botan::DilithiumMode::Dilithium6x5_AES, false));
-         results.push_back(run_roundtrip("Dilithium_8x7_AES", Botan::DilithiumMode::Dilithium8x7_AES, false));
-         results.push_back(run_roundtrip("Dilithium_4x4_AES_Randomized", Botan::DilithiumMode::Dilithium4x4_AES, true));
-         results.push_back(run_roundtrip("Dilithium_6x5_AES_Randomized", Botan::DilithiumMode::Dilithium6x5_AES, true));
-         results.push_back(run_roundtrip("Dilithium_8x7_AES_Randomized", Botan::DilithiumMode::Dilithium8x7_AES, true));
-   #endif
-
-         return results;
+         return {
+            run_roundtrip("Dilithium_4x4_Common", Botan::DilithiumMode::Dilithium4x4, false, 128, 44),
+            run_roundtrip("Dilithium_6x5_Common", Botan::DilithiumMode::Dilithium6x5, false, 192, 65),
+            run_roundtrip("Dilithium_8x7_Common", Botan::DilithiumMode::Dilithium8x7, false, 256, 87),
+            run_roundtrip("Dilithium_4x4_Common_Randomized", Botan::DilithiumMode::Dilithium4x4, true, 128, 44),
+            run_roundtrip("Dilithium_6x5_Common_Randomized", Botan::DilithiumMode::Dilithium6x5, true, 192, 65),
+            run_roundtrip("Dilithium_8x7_Common_Randomized", Botan::DilithiumMode::Dilithium8x7, true, 256, 87),
+            run_roundtrip("Dilithium_4x4_AES", Botan::DilithiumMode::Dilithium4x4_AES, false, 128, 44),
+            run_roundtrip("Dilithium_6x5_AES", Botan::DilithiumMode::Dilithium6x5_AES, false, 192, 65),
+            run_roundtrip("Dilithium_8x7_AES", Botan::DilithiumMode::Dilithium8x7_AES, false, 256, 87),
+            run_roundtrip("Dilithium_4x4_AES_Randomized", Botan::DilithiumMode::Dilithium4x4_AES, true, 128, 44),
+            run_roundtrip("Dilithium_6x5_AES_Randomized", Botan::DilithiumMode::Dilithium6x5_AES, true, 192, 65),
+            run_roundtrip("Dilithium_8x7_AES_Randomized", Botan::DilithiumMode::Dilithium8x7_AES, true, 256, 87),
+            run_roundtrip("ML-DSA_4x4", Botan::DilithiumMode::ML_DSA_4x4, false, 128, 44),
+            run_roundtrip("ML-DSA_6x5", Botan::DilithiumMode::ML_DSA_6x5, false, 192, 65),
+            run_roundtrip("ML-DSA_8x7", Botan::DilithiumMode::ML_DSA_8x7, false, 256, 87),
+            run_roundtrip("ML-DSA_4x4_Randomized", Botan::DilithiumMode::ML_DSA_4x4, true, 128, 44),
+            run_roundtrip("ML-DSA_6x5_Randomized", Botan::DilithiumMode::ML_DSA_6x5, true, 192, 65),
+            run_roundtrip("ML-DSA_8x7_Randomized", Botan::DilithiumMode::ML_DSA_8x7, true, 256, 87),
+         };
       }
 };
 
-BOTAN_REGISTER_TEST("dilithium", "dilithium_roundtrips", DilithiumRoundtripTests);
+BOTAN_REGISTER_TEST("pubkey", "dilithium_roundtrips", DilithiumRoundtripTests);
 
 class Dilithium_Keygen_Tests final : public PK_Key_Generation_Test {
    public:
       std::vector<std::string> keygen_params() const override {
-         return {
-   #if defined(BOTAN_HAS_DILITHIUM_AES)
-            "Dilithium-4x4-AES-r3", "Dilithium-6x5-AES-r3", "Dilithium-8x7-AES-r3",
-   #endif
-   #if defined(BOTAN_HAS_DILITHIUM)
-               "Dilithium-4x4-r3", "Dilithium-6x5-r3", "Dilithium-8x7-r3",
-   #endif
+         std::vector<std::string> all_instances = {
+            "Dilithium-4x4-AES-r3",
+            "Dilithium-6x5-AES-r3",
+            "Dilithium-8x7-AES-r3",
+            "Dilithium-4x4-r3",
+            "Dilithium-6x5-r3",
+            "Dilithium-8x7-r3",
+            "ML-DSA-4x4",
+            "ML-DSA-6x5",
+            "ML-DSA-8x7",
          };
+         std::vector<std::string> available_instances;
+         std::copy_if(all_instances.begin(),
+                      all_instances.end(),
+                      std::back_inserter(available_instances),
+                      [](const std::string& instance) { return Botan::DilithiumMode(instance).is_available(); });
+         return available_instances;
       }
 
-      std::string algo_name() const override { return "Dilithium"; }
+      std::string algo_name(std::string_view param) const override {
+         if(param.starts_with("Dilithium-")) {
+            return "Dilithium";
+         } else {
+            return "ML-DSA";
+         }
+      }
+
+      std::string algo_name() const override { throw Test_Error("No default algo name set for Dilithium"); }
+
+      std::unique_ptr<Botan::Public_Key> public_key_from_raw(std::string_view keygen_params,
+                                                             std::string_view /* provider */,
+                                                             std::span<const uint8_t> raw_pk) const override {
+         return std::make_unique<Botan::Dilithium_PublicKey>(raw_pk, Botan::DilithiumMode(keygen_params));
+      }
 };
 
 BOTAN_REGISTER_TEST("pubkey", "dilithium_keygen", Dilithium_Keygen_Tests);
