@@ -20,10 +20,11 @@
    #include <botan/tls_messages.h>
    #include <botan/tls_version.h>
    #include <botan/internal/loadstor.h>
+   #include <algorithm>
    #include <exception>
+
    #if defined(BOTAN_HAS_TLS_13)
       #include "test_rng.h"
-
       #include <botan/internal/tls_reader.h>
    #endif
 #endif
@@ -206,7 +207,7 @@ class TLS_Message_Parsing_Test final : public Text_Based_Test {
 BOTAN_REGISTER_TEST("tls", "tls_messages", TLS_Message_Parsing_Test);
 
    #if defined(BOTAN_HAS_TLS_13)
-      #if defined(BOTAN_HAS_CURVE_25519)
+      #if defined(BOTAN_HAS_X25519)
 class TLS_Key_Share_CH_Generation_Test final : public Text_Based_Test {
    public:
       TLS_Key_Share_CH_Generation_Test() :
@@ -227,7 +228,7 @@ class TLS_Key_Share_CH_Generation_Test final : public Text_Based_Test {
                                         "\n"
                                         "key_exchange_groups_to_offer = " +
                                         offered_groups);
-         Botan_Tests::Fixed_Output_RNG rng;
+         Fixed_Output_RNG rng;
          rng.add_entropy(rng_data.data(), rng_data.size());
 
          Botan::TLS::Key_Share share(policy, cb, rng);
@@ -255,7 +256,6 @@ class TLS_Extension_Parsing_Test final : public Text_Based_Test {
          const std::vector<uint8_t> protocol = vars.get_opt_bin("Protocol");
          const std::vector<uint8_t> ciphersuite = vars.get_opt_bin("Ciphersuite");
          const std::string exception = vars.get_req_str("Exception");
-         const std::string expected_name = vars.get_opt_str("Name", "");
          const bool is_positive_test = exception.empty();
 
          Test::Result result(extension + " parsing");
